/*******************************************************************************
 * Copyright (c) 2007, 2010 Association for Decentralized Information Management
 * in Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.db.tests.regression.bugs;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import org.junit.Test;
import org.simantics.db.ReadGraph;
import org.simantics.db.Resource;
import org.simantics.db.Session;
import org.simantics.db.WriteGraph;
import org.simantics.db.WriteOnlyGraph;
import org.simantics.db.common.request.DelayedWriteRequest;
import org.simantics.db.common.request.ReadRequest;
import org.simantics.db.common.request.WriteOnlyRequest;
import org.simantics.db.common.request.WriteRequest;
import org.simantics.db.exception.ClusterSetExistException;
import org.simantics.db.exception.DatabaseException;
import org.simantics.db.service.ClusteringSupport;
import org.simantics.db.service.XSupport;
import org.simantics.db.testing.base.ExistingDatabaseTest;
import org.simantics.layer0.Layer0;

/**
 * Tests that cluster set can be used to explicitly control clustering.
 */
public class Issue3420Test1 extends ExistingDatabaseTest {
    private final boolean DEBUG = false;
    private final Vector<Resource> clusterSets = new Vector<Resource>();
    private final Vector<Resource> resources = new Vector<Resource>();
    @Test
    public void test() throws DatabaseException {
        Session session = getSession();
        init();
        session.syncRequest(new Init());
        session.syncRequest(new InitW());
        session.syncRequest(new Check());
        init();
        session.syncRequest(new InitWO());
        session.syncRequest(new Check());
        init();
        session.syncRequest(new InitWD());
        initDWKraaKraa(session);
        session.syncRequest(new Check());
        session.syncRequest(new InitWD2());
        initDWKraaKraa(session);
        session.syncRequest(new Check());
    }
    private void init() {
        resources.clear();
        clusterSets.clear();
        resources.setSize(1<<17);
        clusterSets.setSize(14);
        if (DEBUG)
            System.out.println("Resource count=" + resources.size() + " cluster sets=" + clusterSets.size());
    }
    class Init extends WriteRequest {
        @Override
        public void perform(WriteGraph g) throws DatabaseException {
            Layer0 l0 = Layer0.getInstance(g);
            try {
                g.newClusterSet(l0.InstanceOf);
            } catch (ClusterSetExistException e) {
                // Cluster set exist already. No problem.
            }
        }
    }
    class InitW extends WriteRequest {
        @Override
        public void perform(WriteGraph g) throws DatabaseException {
            Layer0 l0 = Layer0.getInstance(g);
            Resource oldDefault = g.setClusterSet4NewResource(l0.InstanceOf);
            assertTrue(null == oldDefault);
            for (int i=0; i<clusterSets.size(); ++i) {
                Resource r = g.newResource();
                g.newClusterSet(r);
                clusterSets.set(i, r);
            }
            assertTrue(l0.InstanceOf == g.setClusterSet4NewResource(oldDefault));
            for (int i=0; i<resources.size(); ++i) {
                Resource r = clusterSets.get(i%clusterSets.size());
                resources.set(i, g.newResource(r));
            }
        }
    }
    class InitWD extends DelayedWriteRequest {
        @Override
        public void perform(WriteGraph g) throws DatabaseException {
            for (int i=0; i<clusterSets.size(); ++i) {
                Resource r = g.newResource();
                g.newClusterSet(r);
                clusterSets.set(i, r);
            }
            for (int i=0; i<resources.size(); ++i) {
                Resource r = clusterSets.get(i%clusterSets.size());
                resources.set(i, g.newResource(r));
            }
        }
    }
    class InitWD2 extends DelayedWriteRequest {
        @Override
        public void perform(WriteGraph g) throws DatabaseException {
            for (int i=0; i<clusterSets.size(); ++i) {
                Resource r = g.newResource();
                g.newClusterSet(r);
                clusterSets.set(i, r);
            }
            Layer0 l0 = Layer0.getInstance(g);
            for (int i=0; i<resources.size(); ++i) {
                Resource r = clusterSets.get(i%clusterSets.size());
                Resource rr = g.newResource(r);
                g.claim(rr, l0.InstanceOf, l0.Entity);
                g.claimValue(rr, "value");
                resources.set(i, rr);
            }
        }
    }
    private void initDWKraaKraa(Session s)
    throws DatabaseException {
        XSupport xs = s.getService(XSupport.class);
            for (int i=0; i<clusterSets.size(); ++i) {
                Resource r = xs.convertDelayedResourceToResource(clusterSets.get(i));
                clusterSets.set(i, r);
            }
            for (int i=0; i<resources.size(); ++i) {
                Resource r = xs.convertDelayedResourceToResource(resources.get(i));
                resources.set(i, r);
            }
    }
    class InitWO extends WriteOnlyRequest {
        @Override
        public void perform(WriteOnlyGraph g) throws DatabaseException {
            for (int i=0; i<clusterSets.size(); ++i) {
                Resource r = g.newResource();
                g.newClusterSet(r);
                clusterSets.set(i, r);
            }
            for (int i=0; i<resources.size(); ++i) {
                Resource r = clusterSets.get(i%clusterSets.size());
                resources.set(i, g.newResource(r));
            }
        }
    }
    class Check extends ReadRequest {
        @Override
        public void run(ReadGraph g) throws DatabaseException {
            ClusteringSupport cs = g.getService(ClusteringSupport.class);
            Map<Long, Integer> clusters = new HashMap<Long, Integer>();
            for (int i=0; i<resources.size(); ++i){
                long cluster = cs.getCluster(resources.get(i));
                Integer count = clusters.get(cluster);
                if (null == count)
                    count = 0;
                clusters.put(cluster, ++count);
            }
            Set<Long> sortedKeys = new TreeSet<Long>();
            for (Map.Entry<Long, Integer> i : clusters.entrySet())
                sortedKeys.add(i.getKey());
            long sum = 0;
            for (long key : sortedKeys) {
                int count = clusters.get(key);
                if (DEBUG)
                    System.out.println("cluster=" + key + " count=" + count);
                sum += count;
            }
            if (resources.size() != sum) {
                String msg = "Resource count mismatch. expected=" + resources.size() + " realized=" + sum;
                if (DEBUG)
                    System.out.println(msg);
                fail(msg);
            }
        }
    }
}
